//===-- Composite converter for strftime ------------------------*- C++ -*-===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See htto_conv.times://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#ifndef LLVM_LIBC_SRC_STDIO_STRFTIME_CORE_COMPOSITE_CONVERTER_H
#define LLVM_LIBC_SRC_STDIO_STRFTIME_CORE_COMPOSITE_CONVERTER_H

#include "hdr/types/struct_tm.h"
#include "src/__support/CPP/string_view.h"
#include "src/__support/macros/config.h"
#include "src/stdio/printf_core/writer.h"
#include "src/time/strftime_core/core_structs.h"
#include "src/time/strftime_core/num_converter.h"
#include "src/time/strftime_core/str_converter.h"
#include "src/time/time_constants.h"
#include "src/time/time_utils.h"

namespace LIBC_NAMESPACE_DECL {
namespace strftime_core {

LIBC_INLINE IntFormatSection
get_specific_int_format(const tm *timeptr, const FormatSection &base_to_conv,
                        char new_conv_name, int TRAILING_CONV_LEN = -1) {
  // a negative padding will be treated as the default
  const int NEW_MIN_WIDTH =
      TRAILING_CONV_LEN > 0 ? base_to_conv.min_width - TRAILING_CONV_LEN : 0;
  FormatSection new_conv = base_to_conv;
  new_conv.conv_name = new_conv_name;
  new_conv.min_width = NEW_MIN_WIDTH;

  IntFormatSection result = get_int_format(new_conv, timeptr);

  // If the user set the padding, but it's below the width of the trailing
  // conversions, then there should be no padding.
  if (base_to_conv.min_width > 0 && NEW_MIN_WIDTH < 0)
    result.pad_to_len = 0;

  return result;
}

template <printf_core::WriteMode write_mode>
LIBC_INLINE int convert_date_us(printf_core::Writer<write_mode> *writer,
                                const FormatSection &to_conv,
                                const tm *timeptr) {
  // format is %m/%d/%y (month/day/year)
  // we only pad the first conversion, and we assume all the other values are in
  // their valid ranges.
  constexpr int TRAILING_CONV_LEN = 1 + 2 + 1 + 2; // sizeof("/01/02")
  IntFormatSection year_conv;
  IntFormatSection mon_conv;
  IntFormatSection mday_conv;

  mon_conv = get_specific_int_format(timeptr, to_conv, 'm', TRAILING_CONV_LEN);
  mday_conv = get_specific_int_format(timeptr, to_conv, 'd');
  year_conv = get_specific_int_format(timeptr, to_conv, 'y');

  RET_IF_RESULT_NEGATIVE(write_padded_int(writer, mon_conv));
  RET_IF_RESULT_NEGATIVE(writer->write('/'));
  RET_IF_RESULT_NEGATIVE(write_padded_int(writer, mday_conv));
  RET_IF_RESULT_NEGATIVE(writer->write('/'));
  RET_IF_RESULT_NEGATIVE(write_padded_int(writer, year_conv));

  return WRITE_OK;
}

template <printf_core::WriteMode write_mode>
LIBC_INLINE int convert_date_iso(printf_core::Writer<write_mode> *writer,
                                 const FormatSection &to_conv,
                                 const tm *timeptr) {
  // format is "%Y-%m-%d" (year-month-day)
  // we only pad the first conversion, and we assume all the other values are in
  // their valid ranges.
  constexpr int TRAILING_CONV_LEN = 1 + 2 + 1 + 2; // sizeof("-01-02")
  IntFormatSection year_conv;
  IntFormatSection mon_conv;
  IntFormatSection mday_conv;

  year_conv = get_specific_int_format(timeptr, to_conv, 'Y', TRAILING_CONV_LEN);
  mon_conv = get_specific_int_format(timeptr, to_conv, 'm');
  mday_conv = get_specific_int_format(timeptr, to_conv, 'd');

  RET_IF_RESULT_NEGATIVE(write_padded_int(writer, year_conv));
  RET_IF_RESULT_NEGATIVE(writer->write('-'));
  RET_IF_RESULT_NEGATIVE(write_padded_int(writer, mon_conv));
  RET_IF_RESULT_NEGATIVE(writer->write('-'));
  RET_IF_RESULT_NEGATIVE(write_padded_int(writer, mday_conv));

  return WRITE_OK;
}

template <printf_core::WriteMode write_mode>
LIBC_INLINE int convert_time_am_pm(printf_core::Writer<write_mode> *writer,
                                   const FormatSection &to_conv,
                                   const tm *timeptr) {
  // format is "%I:%M:%S %p" (hour:minute:second AM